@@ -1,6 +1,7 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*- */
 
 #include "test-utils.h"
+#include "soup-message-private.h"
 #include "soup-server-message-private.h"
 
 #if HAVE_GNUTLS
@@ -71,6 +72,14 @@ do_strictness_test (gconstpointer data)
 	g_assert_nonnull (soup_message_get_tls_peer_certificate (msg));
 	flags = soup_message_get_tls_peer_certificate_errors (msg);
 
+        if (test->expected_status == SOUP_STATUS_OK) {
+                g_assert_cmpuint (soup_message_get_tls_protocol_version (msg), ==, G_TLS_PROTOCOL_VERSION_TLS_1_3);
+                g_assert_cmpstr (soup_message_get_tls_ciphersuite_name (msg), ==, "TLS_AES-256-GCM_SHA384");
+        } else {
+                g_assert_cmpuint (soup_message_get_tls_protocol_version (msg), ==, G_TLS_PROTOCOL_VERSION_UNKNOWN);
+                g_assert_null (soup_message_get_tls_ciphersuite_name (msg));
+        }
+
 	g_test_bug ("665182");
 	if (test->with_ca_list && !error)
 		g_assert_cmpuint (flags, ==, 0);
@@ -127,7 +136,7 @@ test_tls_interaction_class_init (TestTlsInteractionClass *klass)
 
 
 static gboolean
-accept_client_certificate (GTlsConnection       *server,
+accept_client_certificate (SoupServerMessage    *msg,
 			   GTlsCertificate      *client_cert,
 			   GTlsCertificateFlags  errors)
 {
@@ -136,19 +145,11 @@ accept_client_certificate (GTlsConnection       *server,
 
 static void
 server_request_started (SoupServer        *server,
-			SoupServerMessage *msg,
-			GTlsDatabase      *tls_db)
+                        SoupServerMessage *msg)
 {
-	SoupSocket *sock;
-	GIOStream *conn;
-
-	sock = soup_server_message_get_soup_socket (msg);
-	conn = soup_socket_get_connection (sock);
-	g_tls_connection_set_database (G_TLS_CONNECTION (conn), tls_db);
-	g_object_set (conn, "authentication-mode", G_TLS_AUTHENTICATION_REQUIRED, NULL);
-	g_signal_connect (conn, "accept-certificate",
-			  G_CALLBACK (accept_client_certificate),
-			  NULL);
+        g_signal_connect (msg, "accept-certificate",
+                          G_CALLBACK (accept_client_certificate),
+                          NULL);
 }
 
 static void
@@ -167,10 +168,11 @@ do_tls_interaction_test (gconstpointer data)
 
 	session = soup_test_session_new (NULL);
 	tls_db = soup_session_get_tls_database (session);
+        g_object_set (server, "tls-database", tls_db, "tls-auth-mode", G_TLS_AUTHENTICATION_REQUIRED, NULL);
 
 	g_signal_connect (server, "request-started",
-			  G_CALLBACK (server_request_started),
-			  tls_db);
+                          G_CALLBACK (server_request_started),
+                          session);
 
 	/* Without a GTlsInteraction */
 	msg = soup_message_new_from_uri ("GET", uri);
@@ -203,7 +205,8 @@ do_tls_interaction_test (gconstpointer data)
 	g_bytes_unref (body);
 	g_object_unref (msg);
 
-	g_signal_handlers_disconnect_by_data (server, tls_db);
+        g_object_set (server, "tls-database", NULL, "tls-auth-mode", G_TLS_AUTHENTICATION_NONE, NULL);
+	g_signal_handlers_disconnect_by_data (server, session);
 
 	soup_test_session_abort_unref (session);
 	g_object_unref (certificate);
@@ -313,10 +316,11 @@ do_tls_interaction_msg_test (gconstpointer data)
 
         session = soup_test_session_new (NULL);
         tls_db = soup_session_get_tls_database (session);
+        g_object_set (server, "tls-database", tls_db, "tls-auth-mode", G_TLS_AUTHENTICATION_REQUIRED, NULL);
 
         g_signal_connect (server, "request-started",
                           G_CALLBACK (server_request_started),
-                          tls_db);
+                          session);
 
         /* Not handling request-certificate signal */
         msg = soup_message_new_from_uri ("GET", uri);
@@ -360,6 +364,38 @@ do_tls_interaction_msg_test (gconstpointer data)
         g_bytes_unref (body);
         g_object_unref (msg);
 
+        /* Using the wrong certificate fails */
+        wrong_certificate = g_tls_certificate_new_from_files (
+                g_test_get_filename (G_TEST_DIST, "test-cert-2.pem", NULL),
+                g_test_get_filename (G_TEST_DIST, "test-key-2.pem", NULL),
+                NULL
+        );
+        g_assert_nonnull (wrong_certificate);
+        msg = soup_message_new_from_uri ("GET", uri);
+        soup_message_add_flags (msg, SOUP_MESSAGE_NEW_CONNECTION);
+        g_signal_connect (msg, "request-certificate",
+                          G_CALLBACK (request_certificate_cb),
+                          wrong_certificate);
+        body = soup_test_session_async_send (session, msg, NULL, &error);
+        if (!g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CONNECTION_CLOSED))
+                g_assert_error (error, G_TLS_ERROR, G_TLS_ERROR_CERTIFICATE_REQUIRED);
+        g_assert_null (body);
+        g_clear_error (&error);
+        g_object_unref (msg);
+
+        /* Passing NULL certificate fails */
+        msg = soup_message_new_from_uri ("GET", uri);
+        soup_message_add_flags (msg, SOUP_MESSAGE_NEW_CONNECTION);
+        g_signal_connect (msg, "request-certificate",
+                          G_CALLBACK (request_certificate_cb),
+                          NULL);
+        body = soup_test_session_async_send (session, msg, NULL, &error);
+        if (!g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CONNECTION_CLOSED))
+                g_assert_error (error, G_TLS_ERROR, G_TLS_ERROR_CERTIFICATE_REQUIRED);
+        g_clear_error (&error);
+        g_bytes_unref (body);
+        g_object_unref (msg);
+
         /* request-certificate is not emitted if the certificate is set before the load */
         msg = soup_message_new_from_uri ("GET", uri);
         soup_message_add_flags (msg, SOUP_MESSAGE_NEW_CONNECTION);
@@ -385,19 +421,27 @@ do_tls_interaction_msg_test (gconstpointer data)
         g_object_unref (msg);
 
         /* Using the wrong certificate fails */
-        wrong_certificate = g_tls_certificate_new_from_files (
-                g_test_get_filename (G_TEST_DIST, "test-cert-2.pem", NULL),
-                g_test_get_filename (G_TEST_DIST, "test-key-2.pem", NULL),
-                NULL
-        );
-        g_assert_nonnull (wrong_certificate);
         msg = soup_message_new_from_uri ("GET", uri);
         soup_message_add_flags (msg, SOUP_MESSAGE_NEW_CONNECTION);
         g_signal_connect (msg, "request-certificate",
                           G_CALLBACK (request_certificate_async_cb),
                           wrong_certificate);
         body = soup_test_session_async_send (session, msg, NULL, &error);
-        g_assert_error (error, G_IO_ERROR, G_IO_ERROR_CONNECTION_CLOSED);
+        if (!g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CONNECTION_CLOSED))
+                g_assert_error (error, G_TLS_ERROR, G_TLS_ERROR_CERTIFICATE_REQUIRED);
+        g_assert_null (body);
+        g_clear_error (&error);
+        g_object_unref (msg);
+
+        /* Passing NULL certificate fails */
+        msg = soup_message_new_from_uri ("GET", uri);
+        soup_message_add_flags (msg, SOUP_MESSAGE_NEW_CONNECTION);
+        g_signal_connect (msg, "request-certificate",
+                          G_CALLBACK (request_certificate_async_cb),
+                          NULL);
+        body = soup_test_session_async_send (session, msg, NULL, &error);
+        if (!g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CONNECTION_CLOSED))
+                g_assert_error (error, G_TLS_ERROR, G_TLS_ERROR_CERTIFICATE_REQUIRED);
         g_assert_null (body);
         g_clear_error (&error);
         g_object_unref (msg);
@@ -462,13 +506,186 @@ do_tls_interaction_msg_test (gconstpointer data)
                 g_object_unref (pkcs11_certificate);
         }
 
-        g_signal_handlers_disconnect_by_data (server, tls_db);
+        g_object_set (server, "tls-database", NULL, "tls-auth-mode", G_TLS_AUTHENTICATION_NONE, NULL);
+        g_signal_handlers_disconnect_by_data (server, session);
 
         soup_test_session_abort_unref (session);
         g_object_unref (certificate);
         g_object_unref (wrong_certificate);
 }
 
+static gboolean
+preconnect_request_certificate (SoupMessage          *msg,
+                                GTlsClientConnection *conn,
+                                gboolean             *called)
+{
+        *called = TRUE;
+
+        return FALSE;
+}
+
+static gboolean
+preconnect_request_certificate_password (SoupMessage  *msg,
+                                         GTlsPassword *password,
+                                         gboolean     *called)
+{
+        *called = TRUE;
+
+        return FALSE;
+}
+
+static void
+preconnect_finished_cb (SoupSession  *session,
+                        GAsyncResult *result,
+                        gboolean     *preconnect_finished)
+{
+        g_assert_true (soup_session_preconnect_finish (session, result, NULL));
+        *preconnect_finished = TRUE;
+}
+
+static void
+do_tls_interaction_preconnect_test (gconstpointer data)
+{
+        SoupServer *server = (SoupServer *)data;
+        SoupSession *session;
+        SoupMessage *preconnect_msg;
+        SoupMessage *msg;
+        GBytes *body;
+        GTlsDatabase *tls_db;
+        GTlsCertificate *certificate;
+        GError *error = NULL;
+        gboolean preconnect_request_cert_called = FALSE;
+        gboolean preconnect_request_cert_pass_called = FALSE;
+        gboolean preconnect_finished = FALSE;
+
+        SOUP_TEST_SKIP_IF_NO_TLS;
+
+        session = soup_test_session_new (NULL);
+        tls_db = soup_session_get_tls_database (session);
+        g_object_set (server, "tls-database", tls_db, "tls-auth-mode", G_TLS_AUTHENTICATION_REQUIRED, NULL);
+        g_object_get (server, "tls-certificate", &certificate, NULL);
+        g_signal_connect (server, "request-started",
+                          G_CALLBACK (server_request_started),
+                          session);
+
+        /* Start a preconnect until it get blocked on tls interaction */
+        preconnect_msg = soup_message_new_from_uri ("HEAD", uri);
+        g_signal_connect (preconnect_msg, "request-certificate",
+                          G_CALLBACK (preconnect_request_certificate),
+                          &preconnect_request_cert_called);
+        soup_session_preconnect_async (session, preconnect_msg, G_PRIORITY_DEFAULT, NULL, NULL, NULL);
+        while (!soup_message_has_pending_tls_cert_request (preconnect_msg))
+                g_main_context_iteration (NULL, TRUE);
+
+        /* New message should steal the preconnect connection */
+        msg = soup_message_new_from_uri ("GET", uri);
+        g_signal_connect (msg, "request-certificate",
+                          G_CALLBACK (request_certificate_cb),
+                          certificate);
+        body = soup_test_session_async_send (session, msg, NULL, &error);
+        g_assert_no_error (error);
+        soup_test_assert_message_status (msg, SOUP_STATUS_OK);
+        g_assert_false (preconnect_request_cert_called);
+        g_clear_error (&error);
+        g_bytes_unref (body);
+        g_object_unref (msg);
+        g_object_unref (preconnect_msg);
+
+        soup_session_abort (session);
+
+        /* Preconnect finishes if we set the certificate before */
+        preconnect_msg = soup_message_new_from_uri ("HEAD", uri);
+        g_signal_connect (preconnect_msg, "request-certificate",
+                          G_CALLBACK (preconnect_request_certificate),
+                          &preconnect_request_cert_called);
+        soup_message_set_tls_client_certificate (preconnect_msg, certificate);
+        soup_session_preconnect_async (session, preconnect_msg, G_PRIORITY_DEFAULT, NULL,
+                                       (GAsyncReadyCallback)preconnect_finished_cb,
+                                       &preconnect_finished);
+        while (!preconnect_finished)
+                g_main_context_iteration (NULL, TRUE);
+        g_assert_false (preconnect_request_cert_called);
+        g_object_unref (preconnect_msg);
+        /* New request will use the idle connection without having to provide a certificate */
+        msg = soup_message_new_from_uri ("GET", uri);
+        body = soup_test_session_async_send (session, msg, NULL, &error);
+        g_assert_no_error (error);
+        soup_test_assert_message_status (msg, SOUP_STATUS_OK);
+        g_clear_error (&error);
+        g_bytes_unref (body);
+        g_object_unref (msg);
+
+        soup_session_abort (session);
+
+        /* request-certificate signal is not emitted either if the message stealing the
+         * preconnect connection has a certificate set.
+         */
+        preconnect_msg = soup_message_new_from_uri ("HEAD", uri);
+        g_signal_connect (preconnect_msg, "request-certificate",
+                          G_CALLBACK (preconnect_request_certificate),
+                          &preconnect_request_cert_called);
+        soup_session_preconnect_async (session, preconnect_msg, G_PRIORITY_DEFAULT, NULL, NULL, NULL);
+        while (!soup_message_has_pending_tls_cert_request (preconnect_msg))
+                g_main_context_iteration (NULL, TRUE);
+
+        /* New message should steal the preconnect connection */
+        msg = soup_message_new_from_uri ("GET", uri);
+        soup_message_set_tls_client_certificate (msg, certificate);
+        body = soup_test_session_async_send (session, msg, NULL, &error);
+        g_assert_no_error (error);
+        soup_test_assert_message_status (msg, SOUP_STATUS_OK);
+        g_assert_false (preconnect_request_cert_called);
+        g_clear_error (&error);
+        g_bytes_unref (body);
+        g_object_unref (msg);
+        g_object_unref (preconnect_msg);
+
+        soup_session_abort (session);
+
+        /* Currently on the gnutls backend supports pkcs#11 */
+        if (g_strcmp0 (g_type_name (G_TYPE_FROM_INSTANCE (g_tls_backend_get_default ())), "GTlsBackendGnutls") == 0) {
+                GTlsCertificate *pkcs11_certificate;
+
+                pkcs11_certificate = g_tls_certificate_new_from_pkcs11_uris (
+                        "pkcs11:model=mock;serial=1;token=Mock%20Certificate;id=%4D%6F%63%6B%20%43%65%72%74%69%66%69%63%61%74%65;object=Mock%20Certificate;type=cert",
+                        "pkcs11:model=mock;serial=1;token=Mock%20Certificate;id=%4D%6F%63%6B%20%50%72%69%76%61%74%65%20%4B%65%79;object=Mock%20Private%20Key;type=private",
+                        &error
+                );
+                g_assert_no_error (error);
+
+                preconnect_msg = soup_message_new_from_uri ("HEAD", uri);
+                g_signal_connect (preconnect_msg, "request-certificate-password",
+                                  G_CALLBACK (preconnect_request_certificate_password),
+                                  &preconnect_request_cert_pass_called);
+                soup_message_set_tls_client_certificate (preconnect_msg, pkcs11_certificate);
+                soup_session_preconnect_async (session, preconnect_msg, G_PRIORITY_DEFAULT, NULL, NULL, NULL);
+                while (!soup_message_has_pending_tls_cert_pass_request (preconnect_msg))
+                        g_main_context_iteration (NULL, TRUE);
+
+                /* New message should steal the preconnect connection */
+                msg = soup_message_new_from_uri ("GET", uri);
+                g_signal_connect (msg, "request-certificate-password",
+                                  G_CALLBACK (request_certificate_password_cb),
+                                  "ABC123");
+                body = soup_test_session_async_send (session, msg, NULL, &error);
+                g_assert_no_error (error);
+                soup_test_assert_message_status (msg, SOUP_STATUS_OK);
+                g_assert_false (preconnect_request_cert_pass_called);
+                g_clear_error (&error);
+                g_bytes_unref (body);
+                g_object_unref (msg);
+                g_object_unref (preconnect_msg);
+
+                g_object_unref (pkcs11_certificate);
+        }
+
+        g_object_set (server, "tls-database", NULL, "tls-auth-mode", G_TLS_AUTHENTICATION_NONE, NULL);
+        g_signal_handlers_disconnect_by_data (server, session);
+
+        soup_test_session_abort_unref (session);
+        g_object_unref (certificate);
+}
+
 static void
 server_handler (SoupServer        *server,
 		SoupServerMessage *msg,
@@ -508,6 +725,7 @@ main (int argc, char **argv)
 
 	g_test_add_data_func ("/ssl/tls-interaction", server, do_tls_interaction_test);
         g_test_add_data_func ("/ssl/tls-interaction-msg", server, do_tls_interaction_msg_test);
+        g_test_add_data_func ("/ssl/tls-interaction/preconnect", server, do_tls_interaction_preconnect_test);
 
 	for (i = 0; i < G_N_ELEMENTS (strictness_tests); i++) {
 		g_test_add_data_func (strictness_tests[i].name,
